# coding=utf-8
# Copyright 2022 The OpenBMB team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .config import Config
from .cpm1_config import CPM1Config
from .cpm2_config import CPM2Config
from .cpm3_config import CPM3Config
from .t5_config import T5Config
from .gpt2_config import GPT2Config
from .gptj_config import GPTjConfig
from .bert_config import BertConfig
from .roberta_config import RobertaConfig
from .vit_config import VitConfig
from .longformer_config import LongformerConfig
from .glm_config import GLMConfig
from .opt_config import OPTConfig
